from simple_cdd.exceptions import Fail
from simple_cdd.utils import run_command
import os
import subprocess
import logging

log = logging.getLogger()


class Gnupg:
    """
    Collect all gnupg related functions
    """

    def __init__(self, env):
        self.env = env

    def init_homedir(self):
        gnupghome = self.env.get("GNUPGHOME")
        if not os.path.isdir(gnupghome):
            os.makedirs(gnupghome, exist_ok=True)
            os.chmod(gnupghome, 0o700)

        # Import all keyrings into our gnupg home
        for keyring_file in self.env.get("keyring"):
            if not os.path.exists(keyring_file):
                log.warn("keyring file %s does not exist", keyring_file)
                continue
            self.import_keyring(keyring_file)


    def common_gpg_args(self):
        args = ["gpg", "--batch", "--no-default-keyring"]
        for k in self.env.get("keyring"):
            args.extend(("--keyring", k))
        return args

    def extract_inline_contents(self, pathname, sigpathname):
        args = self.common_gpg_args() + ["--decrypt", sigpathname]
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        if stdout and proc.returncode == 0:
            lines = stdout.splitlines(keepends=True)
            with open(pathname, 'w') as x:
                for line in lines:
                    x.write(line.decode('utf-8'))
        else:
            raise Fail("Unable to extract data from %s to %s, returned %d", sigpathname, pathname, proc.returncode)

    def verify_gpg_sig(self, *extra_args):
        args = self.common_gpg_args()
        args.extend(extra_args)
        retval = run_command("verify gpg signature", args)
        if retval != 0:
            raise Fail("Signature verification failed on %s", args)

    def verify_detached_sig(self, pathname, sigpathname):
        return self.verify_gpg_sig("--verify", sigpathname, pathname)

    def verify_inline_sig(self, pathname):
        return self.verify_gpg_sig("--verify", pathname)

    def import_keyring(self, keyring_file):
        """
        Import a keyring into our keyring file
        """
        env = dict(os.environ)
        env["GNUPGHOME"] = self.env.get("GNUPGHOME")
        proc = subprocess.Popen(["gpg", "--batch", "--import", keyring_file], stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
        stdout, stderr = proc.communicate()
        retval = proc.wait()
        if retval != 0:
            for line in stderr.decode("utf-8").split("\n"):
                log.error("GPG standard error: %s", line)
            raise Fail("Importing %s into %s failed, gpg error code %s", keyring_file, self.env.get("GNUPGHOME"), retval)

    def list_valid_keys(self, keyring_file):
        """
        Generate a sequence of keyIDs for valid signing keys found in the given
        keyring file
        """
        keys_raw = subprocess.check_output(["gpg",
                                            "--batch",
                                            "--no-default-keyring",
                                            "--keyring", keyring_file,
                                            "--list-keys",
                                            "--with-colons"],
                                            universal_newlines=True)
        for line in keys_raw.split("\n"):
            if not line.startswith("pub") and not line.startswith("sub"):
                continue
            fields = line.split(":")
            keyid = fields[4]
            status = fields[11]
            if 'D' in status: continue
            if 's' not in status and 'S' not in status: continue
            # Append '!' to keyid to allow expired keys in the
            # keyring, though reprepro still requires
            # --ignore=expired* arguments in order to treat signatures
            # from expired keys as valid.
            # https://bugs.debian.org/928703
            yield keyid+'!'



from .exceptions import Fail
from debian import deb822
import subprocess
import fcntl
import select
import shlex
import re
import hashlib
import os
import logging
import shutil
import sys

log = logging.getLogger()

if hasattr(shlex, "quote"):
    shell_quote = shlex.quote
else:
    import pipes
    shell_quote = pipes.quote

if hasattr(shutil, "which"):
    shell_which = shutil.which
else:
    # Available since python 1.6:
    # http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    from distutils.spawn import find_executable
    shell_which = find_executable

def stream_output(proc):
    """
    Take a subprocess.Popen object and generate its output, line by line,
    annotated with "stdout" or "stderr". At process termination it generates
    one last element: ("result", return_code) with the return code of the
    process.
    """
    fds = [proc.stdout, proc.stderr]
    bufs = [b"", b""]
    types = ["stdout", "stderr"]
    # Set both pipes as non-blocking
    for fd in fds:
        fcntl.fcntl(fd, fcntl.F_SETFL, os.O_NONBLOCK)
    # Multiplex stdout and stderr with different prefixes
    while len(fds) > 0:
        s = select.select(fds, (), ())
        for fd in s[0]:
            idx = fds.index(fd)
            buf = fd.read()
            if len(buf) == 0:
                fds.pop(idx)
                if len(bufs[idx]) != 0:
                    yield types[idx], bufs.pop(idx)
                types.pop(idx)
            else:
                bufs[idx] += buf
                lines = bufs[idx].split(b"\n")
                bufs[idx] = lines.pop()
                for l in lines:
                    yield types[idx], l
    res = proc.wait()
    yield "result", res


def run_command(name, cmd, env=None, logfd=None):
    """
    Run a command logging its output.

    name is the name used to identify the command in logs

    cmd is the command itself

    env is the environment to run the command in. If missing, the current
    environment is used.

    logfd, if present, is a file where the full output of the command will also
    be written.
    """
    quoted_cmd = " ".join(shell_quote(x) for x in cmd)
    log.debug("%s running command %s", name, quoted_cmd)
    if logfd: print("runcmd:", quoted_cmd, file=logfd)

    # Run the script itself on an empty environment, so that what was
    # documented is exactly what was run
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)

    stderr = []
    for type, val in stream_output(proc):
        if type == "stdout":
            val = val.decode("utf-8").rstrip()
            if logfd: print("stdout:", val, file=logfd)
            log.debug("%s stdout: %s", name, val)
        elif type == "stderr":
            val = val.decode("utf-8").rstrip()
            if logfd: print("stderr:", val, file=logfd)
            if val: stderr.append(val)
            log.debug("%s stderr: %s", name, val)
        elif type == "result":
            if logfd: print("retval:", val, file=logfd)
            log.debug("%s retval: %d", name, val)
            retval = val

    if retval != 0:
        lastlines = min(len(stderr), 5)
        log.error("%s exited with code %s", name, retval)
        log.error("Last %d lines of standard error:", lastlines)
        for line in stderr[-lastlines:]:
            log.error("%s: %s", name, line)

    return retval


def verify_preseed_file(pathname):
    """
    Verify that a preseed file is valid.

    Returns true if it is valid, false otherwise.
    """
    # # verify that preseeding files are valid
    proc = subprocess.Popen(["/usr/bin/debconf-set-selections", "--checkonly", pathname],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    res = None
    re_ignore_stderr = re.compile(r'debconf: DbDriver "passwords" warning: could not open /var/cache/debconf/passwords.dat: Permission denied')
    for type, val in stream_output(proc):
        if type == "stdout":
            pass
        elif type == "stderr":
            val = val.decode("utf-8")
            if not re_ignore_stderr.match(val):
                print(val, file=sys.stderr)
        elif type == "result":
            res = val

    return res == 0


def list_debs(file_or_dir):
    """
    Generated all .deb or .udeb files recursively found inside the given
    directories.

    If an argument is a file, and is a .deb or .udeb, it is also generated
    """
    wanted = re.compile(r".*\.(deb|udeb)$")
    if os.path.isfile(file_or_dir):
        if wanted.search(file_or_dir):
            yield file_or_dir
    elif os.path.isdir(file_or_dir):
        for root, dirs, files in os.walk(file_or_dir):
            for f in files:
                if not wanted.search(f): continue
                yield os.path.join(root, f)
    else:
        log.warn("local package source %s is neither a file nor a directory", file_or_dir)


class Checksums:
    """
    In-memory database of file checksums
    """
    FIELDS = ["MD5Sum", "SHA1", "SHA256"]

    def __init__(self, env):
        # dict mapping pathnames to { size: int, MD5Sum: str, SHA1: str, SHA256: str }
        #
        # Only those checksum fields actually present in the Release file will
        # exist. size is also optional.
        self.by_relname = {}
        # Files that have been parsed to collect checksums
        self.sources = []
        # Parameters
        self.env = env

    def verify_file(self, absname, relname):
        """
        Verify if a file matches the checksums. Fails if it does not, or if we
        do not have a checksum for it.
        """
        # Get the checksum record for this file
        file_sums = self.by_relname.get(relname, None)
        if file_sums is None:
            if self.env.get("ignore_missing_checksums"):
                log.warning("Ignoring missing checksum for %s", relname)
                return
            else:
                raise Fail("No checksums found for %s in %s", relname, ", ".join(self.sources))

        # Check file size if we have it
        expected_size = file_sums.get("size", None)
        if expected_size is not None:
            real_size = os.stat(absname).st_size
            if real_size != expected_size:
                raise Fail("Invalid size for %s: expected %d, got %d", absname, expected_size, real_size)

        # Check the file against the checksums that we have
        for hashtype in self.FIELDS:
            hashsum = file_sums.get(hashtype, None)
            if hashsum is None: continue

            # Compute the checksum
            if hashtype == "MD5SUM":
                hasher = hashlib.md5()
            elif hashtype == "SHA256":
                hasher = hashlib.sha256()
            elif hashtype == "SHA1":
                hasher = hashlib.sha1()
            else:
                continue
            with open(absname, "rb") as fd:
                hasher.update(fd.read())

            # Verify
            if hasher.hexdigest() != hashsum:
                raise Fail("Invalid checksum for %s: expected %s, got %s", absname, hashsum, hasher.hexdigest())

    def parse_release_file(self, pathname):
        """
        Add checksums from a Release file
        """
        self.sources.append(pathname)
        with open(pathname, "rt") as fd:
            for rec in deb822.Deb822.iter_paragraphs(fd, fields=self.FIELDS):
                for hashname in self.FIELDS:
                    hashes = rec.get(hashname, None)
                    if hashes is None: continue
                    for line in hashes.split("\n"):
                        line = line.strip()
                        if not line: continue
                        hashsum, size, pathname = line.split()
                        entry = self.by_relname.get(pathname, None)
                        if entry is None:
                            self.by_relname[pathname] = { "size": int(size), hashname: hashsum }
                        else:
                            entry["size"] = int(size)
                            entry[hashname] = hashsum

    def parse_checksums_file(self, pathname, fieldname):
        """
        Add checksums from a md5sum/sha256sum format file
        """
        self.sources.append(pathname)
        with open(pathname, "rt") as fd:
            for line in fd:
                hashsum, pathname = line.split()
                if pathname.startswith("./"):
                    pathname = pathname[2:]
                entry = self.by_relname.get(pathname, None)
                if entry is None:
                    self.by_relname[pathname] = { fieldname: hashsum }
                else:
                    entry[fieldname] = hashsum





from .env import *
import os
import sys

VARIABLES = [
    PathVar("simple_cdd_dir", os.getcwd(),
            help="simple-cdd working directory"),
    # FIXME: should simple_cdd_dir just be merged with simple_cdd_temp?
    PathVar("simple_cdd_temp", ["{simple_cdd_dir}", "tmp"],
            help="directory where intermediate build data are stored"),
    PathVar("simple_cdd_logs", ["{simple_cdd_temp}", "log"],
            help="directory where execution logs are stored"),
    PathVar("simple_cdd_mirror", ["{simple_cdd_temp}", "mirror"],
            help="directory where the local mirror is stored"),
    PathVar("simple_cdd_basedir", ["{simple_cdd_temp}", "debian-cd"],
            help="working directory for debian-cd"),
    TextVar("server", "deb.debian.org",
            help="default server used for mirrors"),
    ListVar("mirror_components", ["main"],
            help="components to use to build the distributions"),
    ListVar("mirror_files", ["README", "doc/", "tools/"],
            help="extra files or directories to copy from the mirrors to the distribution CD"),
    TextVar("ARCH", Backtick("dpkg", "--print-architecture"),
            help="Architecture of the target distribution"),
    TextVar("CODENAME", Backtick("lsb_release", "--short", "--codename"), cmdline="--dist",
            help="distribution name (jessie, stretch, buster, bullseye, ..., sid)"),
    TextVar("DI_CODENAME", "{CODENAME}",
            help="distribution name for debian-installer"),
    TextVar("DISKTYPE", "CD",
            help="ISO image type ('CD' or 'DVD')"),
    TextVar("locale", cmdline="--locale",
            help="preselect this default locale for the distribution"),
    TextVar("keyboard", cmdline="--keyboard",
            help="preselect this keyboard in debian-installer (console-keymaps-at/keymap debconf question)"),
    BoolVar("proposed_updates", False, "--proposed-updates",
            help="specify if proposed updates should be used"),
    ListVar("kernel_packages", cmdline="--kernel-packages",
            help="list of kernel packages to include on the CD"),
    TextVar("profiles_udeb_dist", cmdline="--profiles-udeb-dist",
            help="pull simple-cdd-profiles udeb from specified distribution"),
    TextVar("extra_udeb_dist", cmdline="--extra-udeb-dist",
            help="pull in udebs from specified distribution"),
    TextVar("security_mirror", "http://security.debian.org/", cmdline="--security-mirror",
            help="security mirror URL"),
    TextVar("backports_mirror", "{debian_mirror}", cmdline="--backports-mirror",
            help="backports mirror URL"),
    ListVar("profiles", cmdline=("--profiles", "-p"),
            help="select profiles (examples in profiles/ dir)"),
    ListVar("local_packages", cmdline="--local-packages",
            help="list of files included on CD's local package repository"),
    ListVar("build_profiles", cmdline=("--build-profiles", "-b"),
            help="profiles only used while building the CD"),
    ListVar("mirror_tools", ["download", "reprepro"], cmdline="--mirror-tools",
            help="list the mirror tools to be used"),
    ListVar("build_tools", "debian-cd", cmdline="--build-tools",
            help="list the build tools to be used"),
    ListVar("auto_profiles", cmdline=["--auto-profiles", "-a"],
            help="automatically install these profiles"),
    TextVar("updates_mirror", "{debian_mirror}", cmdline="--updates-mirror",
            help="specify mirror for $DIST-updates"),
    ListVar("simple_cdd_dirs",
            ["{simple_cdd_dir}", os.path.dirname(os.path.abspath(sys.argv[0])), "/usr/share/simple-cdd"],
            help="directories used to look for simple-cdd support scripts"),
    TextVar("debian_mirror", "http://{server}/debian/", "--debian-mirror",
            help="official Debian mirror to use to get Debian packages"),
    TextVar("rsync_debian_mirror", "{server}::debian",
            help="official Debian mirror to use to get Debian packages in rsync syntax, used only when using the rsync tool"),
    ListVar("mirror_components_extra", "{mirror_components}",
            help="distribution components to use when getting packages from debian_mirror_extra"),
    ListVar("ARCHES", "{ARCH}",
            help="all the architectures that will be present in the built ISO"),
    TextVar("files_debian_mirror", "{debian_mirror}",
            help="Debian mirror base URL that will be used when downloading files"),
    TextVar("simple_cdd_preseed", "preseed/file=/cdrom/simple-cdd/default.preseed",
            help="kernel command line parameter to enable simple-cdd debconf preseeding"),
    PathVar("TASK", ["{simple_cdd_temp}", "simple-cdd.task"],
            # FIXME: documentation to be improved
            help="debian-cd task to use"),
    ListVar("keyring", ["/usr/share/keyrings/debian-archive-keyring.gpg"], cmdline="--keyring",
            help="use specified keyrings (comma-separated) for verifying packages"),
    BoolVar("user_gnupghome", False,
            help="when true, use the user's gnupg keyring instead of a simple-cdd temporary keyring"),
    TextVar("checksum_file_type", "SHA256SUMS",
            help="name of checksum files to use to verify downloaded files"),
    BoolVar("ignore_missing_checksums", False,
            help="when true, don't fail when a file can't be found in a checksum file."),
    BoolVar("do_mirror", True,
            help="when true, build a local mirror"),
    TextVar("di_release", "current",
            help="debian-installer version to use"),
    PathVar("BASEDIR", "{simple_cdd_basedir}",
            help="working directory for debian-cd"),
    PathVar("MIRROR", "{simple_cdd_mirror}",
            help="directory where the local mirror is stored"),
    TextVar("DEBVERSION", None,
            help="Numeric version of Debian release."),
    TextVar("OFFICIAL", "Unofficial",
            help="TODO: passed as is to debian-cd"),
    TextVar("OFFICIAL_VAL", "0",
            help="TODO: passed as is to debian-cd"),
    TextVar("ISOLINUX", "1",
            help="TODO: passed as is to debian-cd"),
    TextVar("NORECOMMENDS", "1",
            help="When set to 1, do not include recommends in the iso image"),
    TextVar("MAXCDS", "1",
            help="TODO: passed as is to debian-cd"),
    TextVar("OMIT_RELEASE_NOTES", "1",
            help="TODO: passed as is to debian-cd"),
    TextVar("DOJIGDO", "0",
            help="TODO: passed as is to debian-cd"),
    TextVar("MAXJIGDOS", "0",
            help="TODO: passed as is to debian-cd"),
    TextVar("INSTALLER_CD", "2",
            help="0: nothing special, "
                 "1: just add debian-installer (use TASK=tasks/debian-installer), "
                 "2: add d-i and base (use TASK=tasks/debian-installer+kernel)"),
    TextVar("MAXISOS", "ALL",
            help="TODO: passed as is to debian-cd"),
    PathVar("OUT", ["{simple_cdd_dir}", "images"],
            help="Directory where the generated ISO images will be stored"),
    ListVar("includes", ["debian-installer+kernel", "debian-installer"],
            # FIXME: I have no idea what this is (enrico)
            help="List of debian-cd include files to add to the current task"),
    TextVar("serial_console_speed", "115200",
            help="serial console speed used when booting with a serial console"),
    TextVar("serial_console_opts", "ttyS0,{serial_console_speed}",
            help="argument passed to the kernel to boot with a serial console"),
    TextVar("debian_cd_dir", "/usr/share/debian-cd",
            help="directory where debian-cd has its tools"),
    BoolVar("require_optional_packages", cmdline="--require-optional-packages",
            help="fail if missing optional packages (*.downloads)"),
    BoolVar("force_preseed", cmdline="--force-preseed",
            help="only issue a warning if a preseeding file is invalid"),
    BoolVar("use_serial_console", cmdline=["--serial-console", "-s"],
            help="enable serial console when booting the ISO image"),
    BoolVar("backports",
            help="use Debian backports"),
    ListVar("backports_packages",
            help="list of packages to get from backports instead of the normal distribution"),
    BoolVar("clean_mirror",
            help="Remove unused packages from the local mirror"),
    BoolVar("vga_normal"),
    ListVar("default_profiles", help="profiles which default to being selected in profile selection menu at install time"),
    ListVar("all_extras", help="files included on the media in the /simple-cdd dir"),
    ListVar("all_packages"),
    ListVar("BASE_INCLUDE"),
    ListVar("upstream_package_lists"),
    ListVar("checksum_files", help="files used to verify checksums on parts of the mirror other than packages, such as the installer media"),
    ListVar("package_files", help="profile specific lists of packages to install"),
    ListVar("preseed_files", help="profile specific preseeding files"),
    ListVar("exclude_files", help="profile specific list of packages to exclude from media"),
    ListVar("KERNEL_PARAMS"),
    ListVar("EXCLUDE"),
    ListVar("verify_release_keys",
            help="key IDs used to verify release files"),
    TextVar("udebs_filter_formula"),
    TextVar("debian_mirror_extra",
            help="if set, also get packages from this mirror"),
    ListVar("reprepro_opts", "-V",
            help="options added to all reprepro invocations"),
    TextVar("debian_cd_emulate_codename", help="if using non-official CODENAME, create links to emulate an existing codename"),
    TextVar("reprepro_retries", help="number of times reprepro attempts to download new dependencies"),
    TextVar("BOOT_TIMEOUT"),
    TextVar("commandline_opts"),
    TextVar("check_not_requested"),
    TextVar("debian_mirror_extra_dist"),
    TextVar("debian_mirror_extra_list_shell_hook"),
    ListVar("extra_udeb_dist_packages"),
    TextVar("mirror_udebcomponents_extra"),
    TextVar("choices"),
    TextVar("default_choices"),
    TextVar("backports_filter_formula"),
    TextVar("SUITE", "stable",
            help="suite used in the target distribution release file (i.e. stable/testing/unstable)"),
    TextVar("GNUPGHOME",
            help="location of the gnupg keyring to use to verify distrbutions"
                 " (overridden to point to a local copy if user_gnupghome is unset)"),
]



from simple_cdd.exceptions import Fail
from simple_cdd.utils import run_command, list_debs, shell_which
from .base import Tool, ToolShell
from debian import deb822
import shutil
import subprocess
import os
import logging

log = logging.getLogger()

@Tool.register
class ToolMirrorReprepro(ToolShell):
    type = "mirror"
    name = "reprepro"

    def run(self):
        """
        Export variables and run the script
        """
        self.check_pre()

        codename = self.env.get("CODENAME")
        di_codename = self.env.get("DI_CODENAME")

        # Distributions in our local mirror
        distributions_used = [codename]
        distributions = {
            codename: {
                "Architectures": "{ARCHES}",
                "Components": "{mirror_components}",
                "UDebComponents": "main",
                "Description": "mirror for {CODENAME}",
                "Suite": "{SUITE}",
            }
        }

        if codename != di_codename:
            distributions_used.append(di_codename)
            distributions[di_codename] = {
                "Architectures": "{ARCHES}",
                "Components": "{mirror_components}",
                "UDebComponents": "main",
                "Description": "mirror for {CODENAME}",
                "Update": "default-udebs",
                "Suite": "{SUITE}",
            }

        # Update methods for our distributions
        updates = {
            "default": {
                "Suite": "*",
                "Architectures": "{ARCHES}",
                "UDebComponents": "",
                "Method": "{debian_mirror}",
                "FilterList": "deinstall package-list",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-base": {
                "Suite": "*",
                "Architectures": "{ARCHES}",
                "UDebComponents": "",
                "Method": "{debian_mirror}",
                "FilterFormula": "priority (== required) | priority (== important) | priority (== standard)",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-udebs": {
                "Suite": "*",
                "Architectures": "{ARCHES}",
                "Components": "",
                "Method": "{debian_mirror}",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-security-legacy": {
                "Suite": "*/updates",
                "Architectures": "{ARCHES}",
                "UDebComponents": "",
                "Method": "{security_mirror}",
                "FilterList": "deinstall package-list",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-base-security-legacy": {
                "Suite": "*/updates",
                "Architectures": "{ARCHES}",
                "UDebComponents": "",
                "Method": "{security_mirror}",
                "FilterFormula": "priority (== required) | priority (== important) | priority (== standard)",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-security": {
                "Suite": "{CODENAME}-security",
                "Architectures": "{ARCHES}",
                "UDebComponents": "",
                "Method": "{security_mirror}",
                "FilterList": "deinstall package-list",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-base-security": {
                "Suite": "{CODENAME}-security",
                "Architectures": "{ARCHES}",
                "UDebComponents": "",
                "Method": "{security_mirror}",
                "FilterFormula": "priority (== required) | priority (== important) | priority (== standard)",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-updates": {
                "Suite": "{CODENAME}-updates",
                "Architectures": "{ARCHES}",
                "UDebComponents": "main",
                "Method": "{updates_mirror}",
                "FilterList": "deinstall package-list",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-base-updates": {
                "Suite": "{CODENAME}-updates",
                "Architectures": "{ARCHES}",
                "UDebComponents": "main",
                "Method": "{updates_mirror}",
                "FilterFormula": "priority (== required) | priority (== important) | priority (== standard)",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-extra": {
                "Suite": "*",
                "Architectures": "{ARCHES}",
                "Components": "{mirror_components_extra}",
                "UDebComponents": "{mirror_udebcomponents_extra}",
                "Method": "{debian_mirror_extra}",
                "FilterList": "deinstall package-list",
                "VerifyRelease": "{verify_release_keys}",
                "ListShellHook": "{debian_mirror_extra_list_shell_hook}",
            },
            "default-proposed-updates": {
                "Suite": "{CODENAME}-proposed-updates",
                "Architectures": "{ARCHES}",
                "UDebComponents": "main",
                "Method": "{debian_mirror}",
                "FilterList": "deinstall package-list",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-base-proposed-updates": {
                "Suite": "{CODENAME}-proposed-updates",
                "Architectures": "{ARCHES}",
                "UDebComponents": "main",
                "Method": "{debian_mirror}",
                "FilterFormula": "priority (== required) | priority (== important) | priority (== standard)",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-profiles-udeb": {
                "Suite": "{profiles_udeb_dist}",
                "Architectures": "{ARCHES}",
                "Components": "",
                "FilterFormula": "package (== simple-cdd-profiles)",
                "Method": "{debian_mirror}",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-extra-udebs": {
                "Suite": "{extra_udeb_dist}",
                "Architectures": "{ARCHES}",
                "Components": "",
                "Method": "{debian_mirror}",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-backports": {
                "Suite": "{CODENAME}-backports",
                "Architectures": "{ARCHES}",
                "UDebComponents": "main",
                "Method": "{backports_mirror}",
                "FilterList": "deinstall backports-packages",
                "VerifyRelease": "{verify_release_keys}",
            },
            "default-backports-filter-formula": {
                "Suite": "{CODENAME}-backports",
                "Architectures": "{ARCHES}",
                "UDebComponents": "main",
                "Method": "{backports_mirror}",
                "FilterFormula": "{backports_filter_formula}",
                "VerifyRelease": "{verify_release_keys}",
            },
        }

        updates_used = ["default", "default-base", "default-udebs"]
        if self.env.get("clean_mirror"):
            # invoke reprepro's "magic delete" rule.
            updates_used.append("-")
        if self.env.get("security_mirror"):
            if codename == "buster" or codename == "stretch" or codename == "jessie":
                updates_used.append("default-base-security-legacy")
                updates_used.append("default-security-legacy")
            else:
                updates_used.append("default-base-security")
                updates_used.append("default-security")
        if self.env.get("debian_mirror_extra"):
            updates_used.append("default-extra")
            if self.env.get("debian_mirror_extra_dist"):
                updates["default-extra"]["Suite"] = "{debian_mirror_extra_dist}"
        if self.env.get("proposed_updates"):
            updates_used.append("default-base-proposed-updates")
            updates_used.append("default-proposed-updates")
        if self.env.get("updates_mirror"):
            updates_used.append("default-base-updates")
            updates_used.append("default-updates")
        if self.env.get("profiles_udeb_dist"):
            updates_used.append("default-profiles-udeb")
        if self.env.get("extra_udeb_dist"):
            updates_used.append("default-extra-udebs")
            filter = []
            for p in self.env.get("extra_udeb_dist_packages"):
                filter.append("package (== {})".format(p))
            updates["default-extra-udebs"]["FilterFormula"] = " | ".join(filter)
        if self.env.get("backports"):
            updates_used.append("default-backports")
            if self.env.get("backports_filter_formula"):
                updates_used.append("default-backports-filter-formula")
        if self.env.get("udebs_filter_formula"):
            updates["default-udebs"]["FilterFormula"] = "{udebs_filter_formula}"

        distributions[codename]["Update"] = " ".join(updates_used)

        confdir = os.path.join(self.env.get("MIRROR"), "conf")
        os.makedirs(confdir, exist_ok=True)

        def print_records(confname, name_field, field_order, records_used, records):
            verify_release_keys = "|".join(self.env.get("verify_release_keys"))
            with open(os.path.join(confdir, confname), "wt") as fd:
                for name in records_used:
                    rec = records[name]
                    print("{}: {}".format(name_field, name), file=fd)
                    for field in field_order:
                        val = rec.get(field, None)
                        if val is None: continue
                        print("{}: {}".format(field, self.env.format(val, verify_release_keys=verify_release_keys)), file=fd)
                    print(file=fd)

        # Generate the conf/distributions file
        field_order = ["Architectures", "Components", "UDebComponents", "Description", "Update", "Suite"]
        print_records("distributions", "Codename", field_order, distributions_used, distributions)

        # Generate the conf/updates file
        field_order = ["Suite", "Architectures", "Components", "UDebComponents", "Method", "FilterList", "FilterFormula", "VerifyRelease", "ListShellHook"]
        print_records("updates", "Name", field_order, updates_used, updates)

        # Generate the conf/backport-packages file
        if self.env.get("backports"):
            with open(os.path.join(confdir, "backports-packages"), "wt") as fd:
                packages = set(self.env.get("backports_packages"))
                for package in sorted(packages):
                    print(package, "install", file=fd)

        # Recreate an empty package-list file
        pkglist_file = os.path.join(confdir, "package-list")
        with open(pkglist_file, "wt") as fd:
            pass

        # Build the environment for running reprepro
        reprepro_env = {}
        for name, val, changed in self.env.export_iter():
            reprepro_env[name] = str(val)

        # include local packages into the mirror
        for src in self.env.get("local_packages"):
            for f in list_debs(src):
                if f.endswith(".deb"):
                    cmd = ["reprepro"]
                    cmd.extend(self.env.get("reprepro_opts"))
                    cmd.extend(["--ignore=wrongdistribution", "includedeb", codename, f])
                    retval = run_command("reprepro: including local deb file", cmd, env=reprepro_env)
                    if retval != 0:
                        raise Fail("reprepro failed with exit code: %d", retval)
                elif f.endswith(".udeb"):
                    cmd = ["reprepro"]
                    cmd.extend(self.env.get("reprepro_opts"))
                    cmd.extend(["--ignore=wrongdistribution", "includeudeb", codename, f])
                    retval = run_command("reprepro: including local deb file", cmd, env=reprepro_env)
                    if retval != 0:
                        raise Fail("reprepro failed with exit code: %d", retval)
                    if codename != di_codename:
                        cmd = ["reprepro"]
                        cmd.extend(self.env.get("reprepro_opts"))
                        cmd.extend(["--ignore=wrongdistribution", "includeudeb", di_codename, f])
                        retval = run_command("reprepro: including local deb file", cmd, env=reprepro_env)
                else:
                    log.warn("unknown package type: %s", f)

        # Update package lists
        cmd = ["reprepro"]
        cmd.extend(self.env.get("reprepro_opts"))
        cmd.extend(["--noskipold", "update"])
        retval = run_command("reprepro: updating package lists", cmd, env=reprepro_env)
        if retval != 0:
            raise Fail("reprepro failed with exit code: %d", retval)

        # TODO: we can compute here via dose-debcheck if $all_packages are all
        # installable, and warn otherwise or fail before we start building a
        # mirror

        self.run_script()

        for f in self.env.get("exclude_files"):
            with open(f, "rt") as infd:
                for line in infd:
                    pkg = line.strip()
                    if not pkg or pkg.startswith("#"): continue
                    # Remove files listed in exclude_files
                    cmd = ["reprepro"]
                    cmd.extend(self.env.get("reprepro_opts"))
                    cmd.extend(["--noskipold", "remove", self.env.get("CODENAME"), pkg])
                    retval = run_command("reprepro: remove console-tools-freebsd", cmd, env=reprepro_env)
                    if retval != 0:
                        raise Fail("reprepro failed with exit code: %d", retval)

        self.check_post(retval)

    def check_post(self, retval):
        debcheck = shell_which("dose-debcheck")
        if debcheck is None:
            log.info("dose-debcheck not found: skipping mirror checks")
            return

        for a in self.env.get("ARCHES"):
            mirror_components = self.env.get("mirror_components")
            for component in mirror_components:

                bg_pkgfile = [
                    self.env.format(
                        "{MIRROR}/dists/{CODENAME}/{component}/binary-{a}/Packages",
                        component=i, a=a)
                    for i in mirror_components if i != component
                ]

                bg_command = []
                for file in bg_pkgfile:
                    bg_command.append('--bg')
                    bg_command.append(file)

                command = [debcheck, "--failures", "--explain"]
                command.extend(bg_command)
                pkgfile = self.env.format(
                    "{MIRROR}/dists/{CODENAME}/{component}/binary-{a}/Packages",
                    component=component, a=a)
                if os.stat(pkgfile).st_size == 0: continue
                log.info("Checking package file %s using %s", pkgfile, debcheck)
                command.append(pkgfile)

                proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = proc.communicate()
                retval = proc.wait()
                if retval != 0:
                    log.warn("Found uninstallable packages in %s:", pkgfile)
                    for line in stdout.decode("utf-8").split("\n"):
                        log.warn("  %s", line)






import logging
import curses
import sys
import shutil
import time

if hasattr(shutil, "get_terminal_size"):
    get_terminal_size = shutil.get_terminal_size
else:
    def get_terminal_size():
        import struct
        import os
        def ioctl_GWINSZ(fd):
            try:
                import fcntl
                import termios
                cr = struct.unpack('hh',
                                fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
                return cr
            except:
                pass
        cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
        if not cr:
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_GWINSZ(fd)
                os.close(fd)
            except:
                pass
        if not cr:
            try:
                cr = (os.environ['LINES'], os.environ['COLUMNS'])
            except:
                return None
        return int(cr[1]), int(cr[0])

class FancyTerminalHandler(logging.Handler):
    """
    A log handler that provides fancy output to a terminal.

    It handles colors, and it turns DEBUG entries into progressbars.

    This is a mix of StreamHandler and tornado's formatter.
    """
    FORMAT = '%(level_color)s%(asctime)-15s %(levelname)s%(end_color)s %(message)s'
    PROGRESS_FORMAT = ' (%(ticker)s) %(levelname)s %(message)s'
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(self, stream=None):
        """
        Initialize the handler.

        If stream is not specified, sys.stderr is used.
        """
        super().__init__()
        self.stream = stream if stream is not None else sys.stderr

        # Log level above which entries are not considered progress entries
        self.non_progress = logging.WARN

        # Build a mapping between level types and color escape sequences
        self._level_colors = {}
        fg_color = (curses.tigetstr("setaf") or
                    curses.tigetstr("setf") or "")
        for levelno, code in (
                (logging.DEBUG,      4),  # Blue
                (logging.INFO,       2),  # Green
                (logging.WARNING,    3),  # Yellow
                (logging.ERROR,      1)): # Red
            self._level_colors[levelno] = str(curses.tparm(fg_color, code), "ascii")
        self._normal = str(curses.tigetstr("sgr0"), "ascii")
        self._progress = str(curses.tparm(fg_color, 6), "ascii") # Cyan
        self._clreol = str(curses.tigetstr("el"), "ascii")
        self._ticker = "-\|/"
        self._ticker_index = 0

        cols, lines = get_terminal_size()
        self._output_width = cols

    def emit(self, record):
        """
        Format and print out a record.
        """
        try:
            record.message = record.getMessage()
            record.asctime = time.strftime(self.DATE_FORMAT, time.localtime(record.created))

            if record.levelno < self.non_progress:
                #record.progress_color = self._progress
                #record.end_color = self._normal
                record.ticker = self._ticker[self._ticker_index]
                self._ticker_index = (self._ticker_index + 1) % len(self._ticker)

                # Progress output
                formatted = (self.PROGRESS_FORMAT % record.__dict__).expandtabs()
                formatted = formatted[:self._output_width]
                stream = self.stream
                stream.write(self._progress)
                stream.write(formatted)
                stream.write(self._normal)
                stream.write(self._clreol)
                stream.write("\r")
            else:
                # Terminal log output
                level_color = self._level_colors.get(record.levelno, None)
                if level_color is not None:
                    record.level_color = level_color
                    record.end_color = self._normal
                else:
                    record.level_color = record.end_color = ''

                formatted = self.FORMAT % record.__dict__

                #if record.exc_info:
                #    if not record.exc_text:
                #        record.exc_text = self.formatException(record.exc_info)
                #if record.exc_text:
                #    lines = "\n".join((formatted.rstrip(), record.exc_text))
                msg = formatted.replace("\n", "\n    ")
                stream = self.stream
                stream.write(msg)
                stream.write(self._clreol)
                stream.write("\n")

            self.flush()
        except Exception:
            self.handleError(record)

    def finish_logging(self):
        """
        Notify that there will be no more logging, so that we can clean things
        up if needed.
        """
        # Send a clear to EOL command: this way, if the last thing that we
        # logged was a progress indicator, we do not have the shell prompt
        # appear in the middle of it
        self.stream.write(self._clreol)
        self.stream.flush()





class Fail(BaseException):
    """
    Exception thrown when some check decided that the CDD building process
    cannot proceed.
    """
    def __str__(self):
        return self.args[0] % self.args[1:]



import unittest
from simple_cdd import env
import tempfile
import os

class TestEnv(unittest.TestCase):
    def setUp(self):
        # Save and clear the environment before the tests
        self.saved_env = dict(os.environ)
        os.environ.clear()

    def tearDown(self):
        # Restore the environment after the tests
        os.environ.clear()
        for k, v in self.saved_env.items():
            os.environ[k] = v

    def test_knownvar_text(self):
        VARIABLES = [
            env.TextVar("test", "default")
        ]

        # If the variable does not exist in the environment, it gets the
        # default value
        e = env.Environment(VARIABLES)
        self.assertEqual(e.test, "default")

        # If the variable does exists in the environment, the environment value
        # is used
        os.environ["test"] = "somevalue"
        e = env.Environment(VARIABLES)
        self.assertEqual(e.test, "somevalue")

    def test_unknownvar(self):
        # Unknown values raise AttributeError via the getattr interface, even
        # if set in the environment, but are still accessible via get()
        e = env.Environment([])
        with self.assertRaises(AttributeError):
            e.test
        self.assertEqual(e.get("test"), "")

        os.environ["test"] = "somevalue"
        e = env.Environment([])
        with self.assertRaises(AttributeError):
            e.test
        self.assertEqual(e.get("test"), "somevalue")

    def test_expansion(self):
        # Case of expansion that broke in the past
        VARIABLES = [
            env.PathVar("simple_cdd_dir", "test",
                    help="simple-cdd working directory"),
            env.PathVar("simple_cdd_temp", ["{simple_cdd_dir}", "tmp"],
                    help="directory where intermediate build data are stored"),
            env.PathVar("simple_cdd_mirror", ["{simple_cdd_temp}", "mirror"],
                    help="directory where the local mirror is stored"),
            env.PathVar("MIRROR", "{simple_cdd_mirror}",
                    help="directory where the local mirror is stored"),
        ]
        e = env.Environment(VARIABLES)
        self.assertEqual(e.MIRROR, "test/tmp/mirror")

        VARIABLES = [
            env.BoolVar("a", True),
            env.BoolVar("b", "{a}"),
        ]
        e = env.Environment(VARIABLES)
        self.assertEqual(e.a, True)

    def test_configfile(self):
        VARIABLES = [
            env.BoolVar("use_security_mirror", help="merge Debian security updates into mirror"),
        ]
        e = env.Environment(VARIABLES)
        self.assertFalse(e.use_security_mirror)
        e.set("use_security_mirror", True)
        self.assertTrue(e.use_security_mirror)

        with tempfile.NamedTemporaryFile(mode="wt") as fd:
            print("use_security_mirror=false", file=fd)
            fd.flush()
            e.read_config_file(fd.name)
        self.assertFalse(e.use_security_mirror)




from simple_cdd.exceptions import Fail
from simple_cdd.utils import run_command, shell_quote
import os.path
import logging

log = logging.getLogger()


class Tool:
    """
    Base class for external tools
    """
    TOOLS = {}

    @classmethod
    def register(cls, tool):
        """
        Decorator used to register a Tool object so that it can be found when
        tools need to be run
        """
        cls.TOOLS[(tool.type, tool.name)] = tool
        return tool

    @classmethod
    def create(cls, env, type, name):
        """
        Create a tool runner for this type and name.
        """
        tool = cls.TOOLS.get((type, name), None)
        if tool is None:
            return ToolShell(env, type, name)
        else:
            return tool(env)


class ToolShell(Tool):
    """
    Base class for tools implemented as shell scripts
    """
    # Tool type (build, mirror, testing)
    type = "unknown"
    # Tool name
    name = "unknown"

    def __init__(self, env, type=None, name=None):
        """
        Basic initialization, and lookup of the full pathname of the shell
        script.

        ToolShell can be used directly without subclassing, to run shell tools
        without extra checks or exports.
        """
        if type is not None: self.type = type
        if name is not None: self.name = name
        self.env = env
        self.pathname = self.find_script()

    def find_script(self):
        """
        Find the full pathname of the script to be run
        """
        for d in self.env.get("simple_cdd_dirs"):
            pathname = os.path.join(d, "tools", self.type, self.name)
            if not os.path.exists(pathname): continue
            return pathname
        raise Fail("Cannot find tool %s/%s in %s", self.type, self.name, self.env.get("simple_cdd_dirs"))

    def check_pre(self):
        """
        Run checks before running the script
        """
        pass

    def check_post(self, retval):
        """
        Run checks after running the script
        """
        pass

    def _make_run_script(self):
        """
        Make a script that can be used to run this tool in the current
        environment
        """
        logdir = self.env.get("simple_cdd_logs")
        scriptname = os.path.join(logdir, "{}-{}".format(self.type, self.name))

        # Write out what we are doing into a script, both as clear
        # documentation and as a way to rerun this step by hand and tweak
        # it for debugging
        with open(scriptname, "wt") as fd:
            print("#!/bin/sh", file=fd)
            print("# Execute tool {}/{}".format(self.type, self.name), file=fd)
            print("", file=fd)
            print("# Export environment", file=fd)
            for name, val, changed in self.env.export_iter():
                if val.help: print("#", val.help, file=fd)
                print("export {}={}".format(name, shell_quote(str(val))), file=fd)
            print("", file=fd)
            print("# Run the tool", file=fd)
            print("exec /bin/sh -ue {} \"$@\"".format(shell_quote(self.pathname)), file=fd)
        os.chmod(scriptname, 0o755)

        return scriptname

    def run_script(self):
        """
        Just build and run the script, without pre and post checks.

        This is useful so that run() can be overridden to do more setup and
        teardown that just the checks, and running the script can still be
        delegated to this method
        """
        # Check if the environment has changed since last run
        log.info("Running tool %s", self.pathname)
        scriptname = self._make_run_script()
        retval = None
        logfilename = scriptname + ".log"
        with open(logfilename, "wt") as fd:
            retval = run_command(
                        "{}/{}".format(self.type, self.name),
                        [scriptname],
                        env={},
                        logfd=fd)

            if retval == 0:
                log.info("%s/%s ran successfully, full log can be found in %s", self.type, self.name, scriptname)
            else:
                raise Fail("%s/%s exited with code %s, full log can be found in %s", self.type, self.name, retval, scriptname)
        return retval

    def run(self):
        """
        Export variables and run the script
        """
        self.check_pre()
        retval = self.run_script()
        self.check_post(retval)






#!/usr/bin/env python3

from setuptools import setup

setup(name="simple-cdd",
      version="0.6.5",
      description="create custom debian-installer CDs",
      long_description="""simple-cdd is a limited though relatively easy tool to create a customized debian-installer CD.

it includes simple mechanisms to create "profiles" that define common system
configurations, which can be selected during system installation. simple-cdd
also makes it easy to build CDs with language and country settings
pre-configured.

this can be used to create a crude "Custom Debian Distribution" using packages
from Debian, with pre-configuration of packages that use debconf.  custom
configuration scripts can be specified to handle packages that don't support
debconf pre-configuration.

testing CD images with qemu is also made simple with a provided script.
""",
      author="Vagrant Cascadian",
      author_email="vagrant@debian.org",
      url="http://anonscm.debian.org/gitweb/?p=collab-maint/simple-cdd.git;a=summary",
      requires=["debian"],
      license="GPL-2",
      platforms="any",
      packages=["simple_cdd", "simple_cdd/tools"],
      # scripts=["build-simple-cdd"],
)



import subprocess
import sys
import os
import re
import json
import tempfile
try:
    # After Python 3.3
    from collections.abc import Iterable
except ImportError:
    # This has changed in Python 3.3 (why, oh why?), reinforcing the idea that
    # the best Python version ever is still 2.7, simply because upstream has
    # promised that they won't touch it (and break it) for at least 5 more
    # years.
    from collections import Iterable
from .utils import shell_quote
import logging

log = logging.getLogger()


class Backtick:
    """
    Represent a value obtained by executing a command and reading its standard
    output. This allows to use commands as default values for configuration
    variables.
    """
    def __init__(self, *command):
        self.val = command
        self.cached_result = None

    def __str__(self):
        return "$({})".format(" ".join(shell_quote(x) for x in self.val))

    def __call__(self):
        if self.cached_result is None:
            self.cached_result = subprocess.check_output(self.val, universal_newlines=True).strip()
        return self.cached_result


class Variable:
    """
    Base class for environment variable proxies.

    This acts as a glue between the configuration files, the environment,
    python, the command line arguments and the scripts that we run.
    """
    def __init__(self, name, default=None, cmdline=None, help=None, env=None):
        # Name of the variable in the environment
        self.name = name
        # Default value of the variable, as a python value that will be read or
        # evaluated when the variable value is requested and none has been set
        self.default = default
        # Name(s) of command line arguments to set this variable
        if isinstance(cmdline, str):
            self.cmdline = [cmdline]
        else:
            self.cmdline = cmdline
        # Documentation for this variable
        self.help = help
        # Current environment (usually set later)
        self.env = env
        # Current value, stored as a string ready to be exported in the
        # environment
        self.current = None
        # Set to true when the variable was automatically created and not
        # defined in the code
        self.automatically_created = False

    def __str__(self):
        """
        Get the current value as a string
        """
        if self.current is None:
            # TODO: build a value from the defaults
            res = self.default_to_string()
        else:
            res = self.current
        return res

    def to_parser(self, parser):
        """
        Add an option for this variable to the command line parser
        """
        default = self.format_default()
        if default == "":
            parser.add_argument(*self.cmdline, action="store", dest=self.name, help=self.help)
        else:
            help = self.help + " (default: {})".format(default)
            parser.add_argument(*self.cmdline, action="store", dest=self.name, help=help)

    def format_default(self):
        """
        Format the default value for showing in documentation and command line
        help
        """
        return str(self.default)

    def default_to_string(self):
        """
        Convert the default value to a string
        """
        if callable(self.default):
            # If 'default' is a python callable, call it and return its result
            return str(self.default())
        elif self.default is None:
            return ""
        elif isinstance(self.default, str):
            return self.env.format(self.default)
        else:
            return str(self.default)

    def to_python(self):
        """
        Return the current value as a python value
        """
        # To be overridden
        return str(self)

    def from_python(self, value):
        """
        Set the current value from a python value
        """
        # To be overridden
        self.current = str(value)

    def from_commandline(self, value):
        """
        Set the current value from a command line option
        """
        if value is None:
            return
        elif value is True:
            self.current = "true"
        elif value is False:
            self.current = ""
        else:
            self.current = value.strip()


class BoolVar(Variable):
    """
    A proxy for a boolean environment variable. For python it behaves as a
    bool, and for the environment it will be "true" for True and "" for False
    """
    def to_parser(self, parser):
        default = self.format_default()
        help = self.help + " (default: {})".format(default)
        parser.add_argument(*self.cmdline, action="store_true", dest=self.name, help=help)

    def format_default(self):
        if isinstance(self.default, bool):
            return "true" if self.default else "false"
        else:
            return super().format_default()

    def default_to_string(self):
        if isinstance(self.default, bool):
            bval = self.default
        else:
            bval = super().default_to_string() == "true"
        return "true" if bval else ""

    def to_python(self):
        return str(self) == "true"

    def from_python(self, value):
        self.current = "true" if value else ""


class TextVar(Variable):
    """
    A proxy for a string environment variable. For python it behaves as a
    str, and for the environment it will be a string.
    """
    pass


class PathVar(Variable):
    """
    A proxy for a string environment variable that contains a path. For python
    it behaves as a str, and for the environment it will be a string. It
    supports appending, and default variables in the form of a string, which
    will be properly concatenated into a path.
    """
    def append(self, value):
        if not isinstance(value, str):
            raise ValueError("Attempted to append a non-string value to a path")
        self.from_python(os.path.join(str(self), value))

    def format_default(self):
        if isinstance(self.default, str):
            return self.default
        elif isinstance(self.default, Iterable):
            return os.path.join(self.default)
        else:
            return super().format_default()

    def default_to_string(self):
        if isinstance(self.default, str):
            return self.env.format(self.default)
        elif isinstance(self.default, Iterable):
            formatted = [self.env.format(x) for x in self.default]
            return os.path.join(*formatted)
        else:
            return super().default_to_string()

class ListVar(Variable):
    """
    A proxy for a string environment variable that contains a list of strings.
    For python it behaves as a sequence, and for the environment it will be a
    space separated string. It supports appending. Command line arguments can
    be space or comma separated.
    """
    def append(self, value):
        cur = self.to_python()
        if isinstance(value, str):
            cur.append(value)
        elif isinstance(value, Iterable):
            cur.extend(value)
        else:
            raise ValueError("Attempted to append a {} value to a list".format(value.__class__))
        self.from_python(cur)

    def format_default(self):
        if isinstance(self.default, str):
            return self.default
        elif isinstance(self.default, Iterable):
            return " ".join(self.default)
        else:
            return super().format_default()

    def default_to_string(self):
        if isinstance(self.default, str):
            return self.env.format(self.default)
        elif isinstance(self.default, Iterable):
            return " ".join(self.env.format(x) for x in self.default)
        else:
            return super().default_to_string()

    def to_python(self):
        val = str(self)
        if val:
            return val.split()
        else:
            return []

    def from_python(self, value):
        if value:
            if isinstance(value, str):
                self.current = value
            elif isinstance(value, Iterable):
                self.current = " ".join(value)
            else:
                raise ValueError("Attempted to append a {} value to a path".format(value.__class__))
        else:
            self.current = ""

    def from_commandline(self, value):
        if isinstance(value, str):
            items = re.split(r"\s*,\s*", value.strip())
            self.current = " ".join(items)
        else:
            super().from_commandline(value)

class Environment:
    """
    Keep track of environment changes
    """
    def __init__(self, variables):
        env = {}
        for v in variables:
            v.env = self
            if v.name in env:
                raise AssertionError("{} defined twice".format(v.name))
            env[v.name] = v
            if v.name in os.environ:
                v.current = os.environ[v.name]
        for name, val in os.environ.items():
            if name in env: continue
            v = TextVar(name, help="Preexisting environment variable {}".format(name))
            env[name] = v
            v.automatically_created = True
            v.current = val
        super().__setattr__("env", env)
        super().__setattr__("initial", dict(os.environ))

    def parse_commandline(self, args):
        """
        Set values from command line arguments
        """
        for name, v in self.env.items():
            if v.cmdline is None: continue
            v.from_commandline(getattr(args, name))

    def get(self, name):
        """
        Get a value by name, returns the empty string if not found
        """
        val = self.env.get(name, None)
        if val is None:
            return ""
        else:
            return val.to_python()

    def set(self, name, value):
        """
        Set a value by name, converting the python value to the type of the variable name.
        """
        cur = self.env.get(name, None)
        if cur is None:
            self.env[name] = cur = TextVar(name, env=self)
            cur.automatically_created = True
        cur.from_python(value)

    def set_from_commandline(self, name, value):
        """
        Set a value by name, converting the user string value to the type of the variable name.
        """
        cur = self.env.get(name, None)
        if cur is None:
            self.env[name] = cur = TextVar(name, env=self)
            cur.automatically_created = True
        cur.from_commandline(value)

    def append(self, name, value):
        """
        Append elements or sequences to a list
        """
        cur = self.env.get(name, None)
        if cur is None:
            self.env[name] = cur = TextVar(name, env=self)
        cur.append(value)

    def unset(self, name):
        """
        Unset an element
        """
        del self.env[name]

    def read_config_file(self, pathname):
        """
        Source a config file and edit configuration from its environment
        """
        log.info("Reading configuration file %s", pathname)
        # Write the source wrapper to a temporary file and execute it
        with tempfile.NamedTemporaryFile("w+t") as fd:
            print("set -a", file=fd)
            print(". " + os.path.abspath(pathname), file=fd)
            print("exec {} -c 'import os,json,sys;json.dump(dict(os.environ),sys.stdout)'".format(sys.executable), file=fd)
            fd.flush()
            out = subprocess.check_output(["sh", fd.name]).decode("utf-8")
        # Parse its output
        if not out.strip(): return
        lines = out.split("\n")
        # Pass through stdout from the script
        for line in lines[:-1]:
            print(line)
        # Parse the environment set by the script
        new_env = json.loads(lines[-1])
        # Diff our environment and the script one, to see what was set
        for key in new_env.keys() - os.environ.keys():
            log.info("%s: new var %s=%s", pathname, key, shell_quote(new_env[key]))
            self.set_from_commandline(key, new_env[key])
        for key in new_env.keys() & os.environ.keys():
            if os.environ[key] == new_env[key]: continue
            log.info("%s: changed var %s=%s", pathname, key, shell_quote(new_env[key]))
            self.set_from_commandline(key, new_env[key])

    def export_iter(self):
        """
        Generate a sequence of (name, value, changed) where:
            name is the variable name
            value is the variable value as a string
            changed is a boolean that is True if the value has changed since
                    the original environmenf at program startup
        """
        for name, val in sorted(self.env.items()):
            valstr = str(val)
            yield name, val, valstr != self.initial.get(name, None)

    def export(self):
        """
        Export variables for the script
        """
        os.environ.clear()
        for name, val, changed in self.export_iter():
            if changed:
                log.info("export %s=%s", name, shell_quote(val))
            #else:
            #    log.info("export %s=%s", name, shell_quote(val))
            os.environ[name] = val

    def format(self, string, *args, **kw):
        """
        Call string.format() including all the known env vars
        """
        kwargs = dict(self.env)
        kwargs.update(**kw)
        return string.format(*args, **kwargs)

    def __getattr__(self, name):
        """
        Get a value by name, raises exception if not found
        """
        val = self.env.get(name, None)
        if val is None or val.automatically_created:
            raise AttributeError("Environment has no variable '{}'".format(name))
        else:
            return val.to_python()

    def __setattr__(self, name, val):
        """
        Get a value by name, raises exception if not found
        """
        val = self.env.get(name, None)
        if val is None:
            super().__setattr__(name, val)
        else:
            cur.from_python(value)



from simple_cdd.exceptions import Fail
from simple_cdd.utils import run_command, Checksums
from simple_cdd.gnupg import Gnupg
from .base import Tool
from urllib.parse import urlparse, urljoin
from urllib import request
import os
import re
import logging

log = logging.getLogger()

@Tool.register
class ToolMirrorDownload(Tool):
    type = "mirror"
    name = "download"

    def __init__(self, env):
        self.env = env
        self.gnupg = Gnupg(env)

    def check_pre(self):
        super().check_pre()
        if not self.env.get("DI_WWW_HOME") and not self.env.get("custom_installer"):
            if not self.env.get("checksum_files"):
                raise Fail("Cannot run mirror/download: checksum_files is empty")
            if not self.env.get("di_match_files"):
                raise Fail("Cannot run mirror/download: di_match_files is empty")

    def run(self):
        env = self.env
        logdir = env.get("simple_cdd_logs")
        logfilename = os.path.join(logdir, "{}-{}.log".format(self.type, self.name))

        with open(logfilename, "wt") as logfd:
            baseurl = env.get("files_debian_mirror")
            path_depth = urlparse(baseurl).path.strip("/").count("/") + 1

            if env.get("http_proxy"):
                os.environ.setdefault('http_proxy', env.get("http_proxy"))

            def _download(url, output, checksums=None, relname=None):
                if checksums:
                    if os.path.exists(output):
                        try:
                            checksums.verify_file(output, relname)
                            log.debug("skipping download: %s checksum matched", output)
                            return
                        except Fail:
                            log.debug("re-downloading: %s checksum invalid", output)
                            pass
                if not os.path.isdir(os.path.dirname(output)):
                    os.makedirs(os.path.dirname(output))
                log.debug("downloading: %s", output)
                request.urlretrieve(url, filename=output)
                if checksums:
                    checksums.verify_file(output, relname)

            if env.get("mirror_files"):
                # Download the checksums present in the archive "extrafiles" and verify
                extrafiles_file_inlinesig = os.path.join(env.get("MIRROR"), "extrafiles")
                extrafiles_file= os.path.join(env.get("simple_cdd_temp"), "extrafiles.unsigned")
                download_extrafiles_file = os.path.join(env.get("files_debian_mirror"), "extrafiles")
                _download(download_extrafiles_file, extrafiles_file_inlinesig)
                self.gnupg.verify_inline_sig(extrafiles_file_inlinesig)
                self.gnupg.extract_inline_contents(extrafiles_file, extrafiles_file_inlinesig)

                # import checksums
                extrafile_sums = Checksums(self.env)
                extrafile_sums.parse_checksums_file(extrafiles_file, 'SHA256')

                with open(extrafiles_file, 'r') as ef:
                    efile = ef.readlines()
                match_mirror_files = []
                for m in env.get("mirror_files"):
                    if m.endswith('/'):
                        match_mirror_files.append(re.escape(m))
                    else:
                        match_mirror_files.append(re.escape(m) + "$")
                match_mirror_files = "(" + "|".join(match_mirror_files) + ")"
                ef_match = re.compile(match_mirror_files)
                ef_files = []
                for line in efile:
                    hashsum, relname = line.split()
                    if ef_match.match(relname):
                        ef_files.append({
                            "absname": os.path.join(env.get("MIRROR"), relname),
                            "relname": relname,
                            "url": os.path.join(env.get("files_debian_mirror"), relname),
                        })

                for x in ef_files:
                    _download(x["url"], x["absname"], checksums=extrafile_sums, relname=x["relname"])

            checksum_files = env.get("checksum_files")

            # Download files needed to build debian-installer image
            files = []
            files.extend(checksum_files)

            if checksum_files:
                # Get the release file and verify that it is valid
                release_file = os.path.join(env.get("simple_cdd_temp"), env.format("{DI_CODENAME}_Release"))
                download_release_file = os.path.join(env.get("files_debian_mirror"), "dists", env.get("DI_CODENAME"), "Release")
                _download(download_release_file, release_file)
                _download(download_release_file + ".gpg", release_file + ".gpg")
                self.gnupg.verify_detached_sig(release_file, release_file + ".gpg")

                # Parse the release file for checksums
                sums = Checksums(self.env)
                sums.parse_release_file(release_file)

                # Ensure that the checksum files are those referenced in the Release file
                # And build a list of additional files to download, matching
                # di_match_files in the checksum files contents
                di_match = re.compile(env.get("di_match_files"))
                for file in checksum_files:
                    if file.endswith("SHA256SUMS"):
                        hashtype = "SHA256"
                    elif file.endswith("MD5SUMS"):
                        hashtype = "MD5Sum"
                    else:
                        log.warn("Unknown hash type for %s, skipping file", file)
                        continue

                    separator = os.path.join('dists/', env.get("DI_CODENAME"), '')
                    separator, relname = file.split(separator)
                    absname = os.path.join(env.get("MIRROR"), file)
                    url = os.path.join(env.get("files_debian_mirror"), file)
                    # Validate the file
                    _download(url, absname, checksums=sums, relname=relname)

                    # Get the list of extra files to download: those whose
                    # pathname matches di_match
                    dirname = os.path.dirname(file)
                    extra_files = []
                    with open(absname, "rt") as fd:
                        for line in fd:
                            hashsum, relname = line.split()
                            if not di_match.search(relname): continue
                            if relname.startswith("./"): relname = relname[2:]
                            extra_files.append({
                                "absname": os.path.join(env.get("MIRROR"), dirname, relname),
                                "relname": relname,
                                "url": os.path.join(env.get("files_debian_mirror"), dirname, relname),
                            })

                    # Check downloaded files against their corresponding checksums.
                    file_sums = Checksums(self.env)
                    file_sums.parse_checksums_file(absname, hashtype)
                    for f in extra_files:
                        # Download the extra files
                        _download(f["url"], f["absname"], checksums=file_sums, relname=f["relname"])


